"""
Typing definitions for thelper.
"""

import io
from typing import TYPE_CHECKING, Any, AnyStr, Callable, Dict, List, Optional, Tuple, Union  # noqa: F401

import matplotlib.pyplot as plt
import numpy as np
from torch import Tensor

if TYPE_CHECKING:
    from thelper.tasks.utils import Task
    from thelper.nn.utils import Module
    from thelper.data.loaders import DataLoader

    ArrayType = np.ndarray
    ArrayShapeType = Union[List[int], Tuple[int]]
    OneOrManyArrayType = Union[List[ArrayType], ArrayType]
    LabelColorMapType = Union[ArrayType, Dict[int, ArrayType]]
    LabelIndex = AnyStr
    LabelType = AnyStr
    LabelDict = Dict[LabelIndex, LabelType]
    LabelList = List[LabelType]
    DrawingType = Union[Tuple[plt.Figure, plt.Axes], None]

    # iteration callbacks should have the following signature:
    #   func(sample, pred, iter_idx, max_iters, epoch_idx, max_epochs)
    SampleType = Dict[Union[AnyStr, int], Tensor]
    PredictionType = Tensor
    IterCallbackType = Optional[Callable[[SampleType, Task, PredictionType, int, int, int, int], None]]

    ConfigIndex = AnyStr
    ConfigValue = Union[AnyStr, bool, float, int, List[Any], Dict[Any, Any]]
    ConfigDict = Dict[ConfigIndex, ConfigValue]

    CheckpointLoadingType = Union[AnyStr, io.FileIO]
    CheckpointContentType = Dict[AnyStr, Any]
    MapLocationType = Union[Callable, AnyStr, Dict[AnyStr, AnyStr]]

    ModelType = Module
    LoaderType = DataLoader
    MultiLoaderType = List[Union[LoaderType, None]]

IterCallbackParams = ["sample", "task", "pred", "iter_idx", "max_iters", "epoch_idx", "max_epochs"]



# noinspection PyPackageRequirements
import torch

import thelper
import thelper.optim.schedulers


def test_transform_custom_step_schedulers():
    optim = torch.optim.SGD(torch.nn.Conv2d(5, 10, 3).parameters(), lr=777)
    schedulers = thelper.optim.CustomStepLR(optim, milestones={
        0: 1/16,
        2: 1/8,
        4: 1/4,
        8: 1/3,
        12: 2/5,
        15: 1/2
    }), thelper.optim.CustomStepLR(optim, milestones={
        2: 1 / 8,
        4: 1 / 4,
        8: 1 / 3,
        12: 2 / 5,
        15: 1 / 2
    })
    print("start")
    for scheduler in schedulers:
        print("sched = %s" % str(scheduler))
        for epoch in range(100):
            print("epoch = %s" % str(epoch))
            scheduler.step(epoch)
            closest_k = None
            for k, v in scheduler.milestones.items():
                if k <= epoch:
                    closest_k = k
                else:
                    break
            sched_lr = scheduler.get_lr()
            if closest_k is None:
                assert sched_lr == [777]
            else:
                print("epoch = %d" % epoch)
                print("closest = " + str(scheduler.milestones[closest_k]))

                assert sched_lr == [777 * scheduler.milestones[closest_k]]
            if epoch >= 18:
                assert sched_lr == [777 * (1 / 2)]



"""Transformations utilities module.

This module contains utility functions used to instantiate transformation/augmentation ops.
"""

import logging

import torchvision.transforms
import torchvision.utils

import thelper.utils

logger = logging.getLogger(__name__)


def load_transforms(stages):
    """Loads a transformation pipeline from a list of stages.

    Each entry in the provided list will be considered a stage in the pipeline. The ordering of the stages
    is important, as some transformations might not be compatible if taken out of order. The entries must
    each be dictionaries that define an operation, its parameters, and some meta-parameters (detailed below).

    The ``operation`` field of each stage will be used to dynamically import a specific type of operation to
    apply. The ``params`` field of each stage will then be used to pass parameters to the constructor of
    this operation.

    If an operation is identified as ``"Augmentor.Pipeline"`` or ``"albumentations.Compose"``, it will be
    specially handled. In both case, the ``params`` field becomes mandatory in the stage dictionary, and it
    must specify the Augmentor or albumentations pipeline operation names and parameters (as a dictionary).
    Two additional optional config fields can then be set for Augmentor pipelines: ``input_tensor`` (bool)
    which specifies whether the previous stage provides a ``torch.Tensor`` to the pipeline (default=False);
    and ``output_tensor`` (bool) which specifies whether the output of the pipeline should be converted into
    a tensor (default=False). For albumentations pipelines, two additional fields are also available, namely
    ``bbox_params`` (dict) and ``keypoint_params`` (dict). For more information on these, refer to the
    documentation of ``albumentations.core.composition.Compose``. Finally, when unpacking dictionaries for
    albumentations pipelines, the keys associated to bounding boxes/masks/keypoints that must be forwarded
    to the composer can be specified via the ``bboxes_key``, ``mask_key``, and ``keypoints_key`` fields.

    All operations can also specify which sample components they should be applied to via the ``target_key``
    field. This field can contain a single key (typically a string), or a list of keys. The operation will
    be applied at runtime to all values which are found in the samples with one of those keys. If no key is
    provided for an operation, it will be applied to all array-like components of the sample. Finally, all
    operations can specify a ``linked_fate`` field (bool) to specify whether the samples provided in lists
    should all have the same fate or not (default=True).

    Usage examples inside a session configuration file::

        # ...
        # the 'loaders' field may contain several transformation pipelines
        # (see 'thelper.data.utils.create_loaders' for more information on these pipelines)
        "loaders": {
            # ...
            # the 'base_transforms' operations are applied to all loaded samples
            "base_transforms": [
                {
                    "operation": "...",
                    "params": {
                        ...
                    },
                    "target_key": [ ... ],
                    "linked_fate": ...
                },
                {
                    "operation": "...",
                    "params": {
                        ...
                    },
                    "target_key": [ ... ],
                    "linked_fate": ...
                },
                ...
            ],
        # ...

    Args:
        stages: a list defining a series of transformations to apply as a single pipeline.

    Returns:
        A transformation pipeline object compatible with the ``torchvision.transforms`` interface.

    .. seealso::
        | :class:`thelper.transforms.wrappers.AlbumentationsWrapper`
        | :class:`thelper.transforms.wrappers.AugmentorWrapper`
        | :class:`thelper.transforms.wrappers.TransformWrapper`
        | :func:`thelper.transforms.utils.load_augments`
        | :func:`thelper.data.utils.create_loaders`
    """
    if not isinstance(stages, list):
        raise AssertionError("expected stages to be provided as a list")
    if not stages:
        return None, True  # no-op transform, and dont-care append
    if not all([isinstance(stage, dict) for stage in stages]):
        raise AssertionError("expected all stages to be provided as dictionaries")
    operations = []
    for stage_idx, stage in enumerate(stages):
        if "operation" not in stage or not stage["operation"]:
            raise AssertionError("stage #%d is missing its operation field" % stage_idx)
        operation_name = stage["operation"]
        operation_params = thelper.utils.get_key_def(["params", "parameters"], stage, {})
        if not isinstance(operation_params, dict):
            raise AssertionError("stage #%d parameters are not provided as a dictionary" % stage_idx)
        operation_targets = None
        if "target_key" in stage:
            if not isinstance(stage["target_key"], (list, str, int)):
                raise AssertionError("stage #%d target keys are not provided as a list or string/int" % stage_idx)
            operation_targets = stage["target_key"] if isinstance(stage["target_key"], list) else [stage["target_key"]]
        linked_fate = thelper.utils.str2bool(stage["linked_fate"]) if "linked_fate" in stage else True
        if operation_name == "Augmentor.Pipeline":
            import Augmentor
            pipeline = Augmentor.Pipeline()
            if not isinstance(operation_params, dict) or not operation_params:
                raise AssertionError("augmentor pipeline 'params' field should contain dictionary of suboperations")
            for pipeline_op_name, pipeline_op_params in operation_params.items():
                getattr(pipeline, pipeline_op_name)(**pipeline_op_params)
            if "input_tensor" in stage and thelper.utils.str2bool(stage["input_tensor"]):
                operations.append(torchvision.transforms.ToPILImage())
            operations.append(thelper.transforms.wrappers.AugmentorWrapper(pipeline, operation_targets, linked_fate))
            if "output_tensor" in stage and thelper.utils.str2bool(stage["output_tensor"]):
                operations.append(torchvision.transforms.ToTensor())
        elif operation_name == "albumentations.Compose":
            if not isinstance(operation_params, dict) or not operation_params:
                raise AssertionError("albumentations pipeline 'params' field should contain dictionary of suboperations")
            suboperations = []
            for op_name, op_params in operation_params.items():
                if not op_name.startswith("albumentations."):
                    op_name = "albumentations." + op_name
                op_type = thelper.utils.import_class(op_name)
                suboperations.append(op_type(**op_params))
            probability = thelper.utils.get_key_def("probability", stage, 1.0)
            to_tensor = thelper.utils.get_key_def("to_tensor", stage, None)
            bbox_params = thelper.utils.get_key_def("bbox_params", stage, {})
            add_targets = thelper.utils.get_key_def("add_targets", stage, {})
            bboxes_key = thelper.utils.get_key_def("bboxes_key", stage, "bbox")
            mask_key = thelper.utils.get_key_def("mask_key", stage, "mask")
            keypoints_key = thelper.utils.get_key_def("keypoints_key", stage, "keypoints")
            cvt_kpts_to_bboxes = thelper.utils.str2bool(thelper.utils.get_key_def("cvt_kpts_to_bboxes", stage, False))
            operations.append(thelper.transforms.wrappers.AlbumentationsWrapper(
                transforms=suboperations, to_tensor=to_tensor, bbox_params=bbox_params, add_targets=add_targets,
                image_key=operation_targets, bboxes_key=bboxes_key, mask_key=mask_key, keypoints_key=keypoints_key,
                probability=probability, cvt_kpts_to_bboxes=cvt_kpts_to_bboxes, linked_fate=linked_fate))
        else:
            operation_type = thelper.utils.import_class(operation_name)
            operation = operation_type(**operation_params)
            if not isinstance(operation, (thelper.transforms.wrappers.TransformWrapper,
                                          thelper.transforms.operations.NoTransformWrapper,
                                          torchvision.transforms.Compose)):
                operations.append(thelper.transforms.wrappers.TransformWrapper(operation,
                                                                               target_keys=operation_targets,
                                                                               linked_fate=linked_fate))
            else:
                operations.append(operation)
    if len(operations) > 1:
        return thelper.transforms.Compose(operations)
    elif len(operations) == 1:
        return operations[0]
    else:
        return None


def load_augments(config):
    """Loads a data augmentation pipeline.

    An augmentation pipeline is essentially a specialized transformation pipeline that can be appended or
    prefixed to the base transforms defined for all samples. Augmentations are typically used to diversify
    the samples within the training set in order to help model generalization. They can also be applied to
    validation and test samples in order to get multiple responses for the same input so that they can
    be averaged/concatenated into a single output.

    Usage examples inside a session configuration file::

        # ...
        # the 'loaders' field can contain several augmentation pipelines
        # (see 'thelper.data.utils.create_loaders' for more information on these pipelines)
        "loaders": {
            # ...
            # the 'train_augments' operations are applied to training samples only
            "train_augments": {
                # specifies whether to apply the augmentations before or after the base transforms
                "append": false,
                "transforms": [
                    {
                        # here, we use a single stage, which is actually an augmentor sub-pipeline
                        # that is purely probabilistic (i.e. it does not increase input sample count)
                        "operation": "Augmentor.Pipeline",
                        "params": {
                            # the augmentor pipeline defines two operations: rotations and flips
                            "rotate_random_90": {"probability": 0.75},
                            "flip_random": {"probability": 0.75}
                        }
                    }
                ]
            },
            # ...
        }
        # ...

    Args:
        config: the configuration dictionary defining the meta parameters as well as the list of transformation
            operations of the augmentation pipeline.

    Returns:
        A tuple that consists of a pipeline compatible with the ``torchvision.transforms`` interfaces, and
        a bool specifying whether this pipeline should be appended or prefixed to the base transforms.

    .. seealso::
        | :class:`thelper.transforms.wrappers.AugmentorWrapper`
        | :func:`thelper.transforms.utils.load_transforms`
        | :func:`thelper.data.utils.create_loaders`
    """
    if not isinstance(config, dict):
        raise AssertionError("augmentation config should be provided as dictionary")
    augments = None
    augments_append = False
    if "append" in config:
        augments_append = thelper.utils.str2bool(config["append"])
    if "transforms" in config and config["transforms"]:
        augments = thelper.transforms.load_transforms(config["transforms"])
    return augments, augments_append



# noinspection PyPackageRequirements
import mock

import thelper
import thelper.utils as tu


# noinspection PyUnusedLocal
@mock.patch('thelper.utils.get_available_cuda_devices', return_value=[])    # no device available
@mock.patch('torch.load', return_value={'version': thelper.__version__})
def test_load_checkpoint_default_cpu_no_devices(mock_torch_load, mock_thelper_devices):
    checkpoint_file = 'dummy-checkpoint.pth'
    tu.load_checkpoint(checkpoint_file)
    mock_torch_load.assert_called_once_with(checkpoint_file, map_location='cpu')


# noinspection PyUnusedLocal
@mock.patch('thelper.utils.get_available_cuda_devices', return_value=[0])   # one device available
@mock.patch('torch.load', return_value={'version': thelper.__version__})
def test_load_checkpoint_not_default_cpu_with_devices(mock_torch_load, mock_thelper_devices):
    checkpoint_file = 'dummy-checkpoint.pth'
    tu.load_checkpoint(checkpoint_file)
    mock_torch_load.assert_called_once_with(checkpoint_file, map_location=None)



"""Graphical User Interface (GUI) utility module.

This module contains various tools and utilities used to instantiate annotators and GUI elements.
"""
import logging

import thelper.utils

logger = logging.getLogger(__name__)


def create_annotator(session_name, save_dir, config, datasets):
    """Instantiates a GUI annotation tool based on the type contained in the config dictionary.

    The tool type is expected to be in the configuration dictionary's `annotator` field, under the `type` key. For more
    information on the configuration, refer to :class:`thelper.gui.annotators.Annotator`. The instantiated type must be
    compatible with the constructor signature of :class:`thelper.gui.annotators.Annotator`. The object's constructor will
    be given the full config dictionary.

    Args:
        session_name: name of the annotation session used for printing and to create output directories.
        save_dir: path to the session directory where annotations and other outputs will be saved.
        config: full configuration dictionary that will be parsed for annotator parameters.
        datasets: map of named dataset parsers that will provide the data to annotate.

    Returns:
        The fully-constructed annotator object, ready to begin annotation via its ``run()`` function.

    .. seealso::
        | :class:`thelper.gui.annotators.Annotator`

    """
    if "annotator" not in config or not config["annotator"]:
        raise AssertionError("config missing 'annotator' field")
    annotator_config = config["annotator"]
    if "type" not in annotator_config or not annotator_config["type"]:
        raise AssertionError("annotator config missing 'type' field")
    annotator_type = thelper.utils.import_class(annotator_config["type"])
    return annotator_type(session_name, config, save_dir, datasets)



"""Dataset utility functions and tools.

This module contains utility functions and tools used to instantiate data loaders and parsers.
"""

import json
import logging
import os
import sys

import thelper.tasks
import thelper.transforms
import thelper.utils

logger = logging.getLogger(__name__)


def create_loaders(config, save_dir=None):
    """Prepares the task and data loaders for a model trainer based on a provided data configuration.

    This function will parse a configuration dictionary and extract all the information required to
    instantiate the requested dataset parsers. Then, combining the task metadata of all these parsers, it
    will evenly split the available samples into three sets (training, validation, test) to be handled by
    different data loaders. These will finally be returned along with the (global) task object.

    The configuration dictionary is expected to contain two fields: ``loaders``, which specifies all
    parameters required for establishing the dataset split, shuffling seeds, and batch size (these are
    listed and detailed below); and ``datasets``, which lists the dataset parser interfaces to instantiate
    as well as their parameters. For more information on the ``datasets`` field, refer to
    :func:`thelper.data.utils.create_parsers`.

    The parameters expected in the 'loaders' configuration field are the following:

    - ``<train_/valid_/test_>batch_size`` (mandatory): specifies the (mini)batch size to use in data
      loaders. If you get an 'out of memory' error at runtime, try reducing it.
    - ``<train_/valid_/test_>collate_fn`` (optional): specifies the collate function to use in data
      loaders. The default one is typically fine, but some datasets might require a custom function.
    - ``shuffle`` (optional, default=True): specifies whether the data loaders should shuffle
      their samples or not.
    - ``test_seed`` (optional): specifies the RNG seed to use when splitting test data. If no seed
      is specified, the RNG will be initialized with a device-specific or time-related seed.
    - ``valid_seed`` (optional): specifies the RNG seed to use when splitting validation data. If no
      seed is specified, the RNG will be initialized with a device-specific or time-related seed.
    - ``torch_seed`` (optional): specifies the RNG seed to use for torch-related stochastic operations
      (e.g. for data augmentation). If no seed is specified, the RNG will be initialized with a
      device-specific or time-related seed.
    - ``numpy_seed`` (optional): specifies the RNG seed to use for numpy-related stochastic operations
      (e.g. for data augmentation). If no seed is specified, the RNG will be initialized with a
      device-specific or time-related seed.
    - ``random_seed`` (optional): specifies the RNG seed to use for stochastic operations with python's
      'random' package. If no seed is specified, the RNG will be initialized with a device-specific or
      time-related seed.
    - ``workers`` (optional, default=1): specifies the number of threads to use to preload batches in
      parallel; can be 0 (loading will be on main thread), or an integer >= 1.
    - ``pin_memory`` (optional, default=False): specifies whether the data loaders will copy tensors
      into CUDA-pinned memory before returning them.
    - ``drop_last`` (optional, default=False): specifies whether to drop the last incomplete batch
      or not if the dataset size is not a multiple of the batch size.
    - ``sampler`` (optional): specifies a type of sampler and its constructor parameters to be used
      in the data loaders. This can be used for example to help rebalance a dataset based on its
      class distribution. See :mod:`thelper.data.samplers` for more information.
    - ``augments`` (optional): provides a list of transformation operations used to augment all samples
      of a dataset. See :func:`thelper.transforms.utils.load_augments` for more info.
    - ``train_augments`` (optional): provides a list of transformation operations used to augment the
      training samples of a dataset. See :func:`thelper.transforms.utils.load_augments` for more info.
    - ``valid_augments`` (optional): provides a list of transformation operations used to augment the
      validation samples of a dataset. See :func:`thelper.transforms.utils.load_augments` for more info.
    - ``test_augments`` (optional): provides a list of transformation operations used to augment the
      test samples of a dataset. See :func:`thelper.transforms.utils.load_augments` for more info.
    - ``eval_augments`` (optional): provides a list of transformation operations used to augment the
      validation and test samples of a dataset. See :func:`thelper.transforms.utils.load_augments` for more info.
    - ``base_transforms`` (optional): provides a list of transformation operations to apply to all
      loaded samples. This list will be passed to the constructor of all instantiated dataset parsers.
      See :func:`thelper.transforms.utils.load_transforms` for more info.
    - ``train_split`` (optional): provides the proportion of samples of each dataset to hand off to the
      training data loader. These proportions are given in a dictionary format (``name: ratio``).
    - ``valid_split`` (optional): provides the proportion of samples of each dataset to hand off to the
      validation data loader. These proportions are given in a dictionary format (``name: ratio``).
    - ``test_split`` (optional): provides the proportion of samples of each dataset to hand off to the
      test data loader. These proportions are given in a dictionary format (``name: ratio``).
    - ``skip_verif`` (optional, default=True): specifies whether the dataset split should be verified
      if resuming a session by parsing the log files generated earlier.
    - ``skip_split_norm`` (optional, default=False): specifies whether the question about normalizing
      the split ratios should be skipped or not.
    - ``skip_class_balancing`` (optional, default=False): specifies whether the balancing of class
      labels should be skipped in case the task is classification-related.

    Example configuration file::

        # ...
        "loaders": {
            "batch_size": 128,  # batch size to use in data loaders
            "shuffle": true,  # specifies that the data should be shuffled
            "workers": 4,  # number of threads to pre-fetch data batches with
            "sampler": {  # we can use a data sampler to rebalance classes (optional)
                # see e.g. 'thelper.data.samplers.WeightedSubsetRandomSampler'
                # ...
            },
            "train_augments": { # training data augmentation operations
                # see 'thelper.transforms.utils.load_augments'
                # ...
            },
            "eval_augments": { # evaluation (valid/test) data augmentation operations
                # see 'thelper.transforms.utils.load_augments'
                # ...
            },
            "base_transforms": { # global sample transformation operations
                # see 'thelper.transforms.utils.load_transforms'
                # ...
            },
            # finally, we define a 80%-10%-10% split for our data
            # (we could instead use one dataset for training and one for testing)
            "train_split": {
                "dataset_A": 0.8
                "dataset_B": 0.8
            },
            "valid_split": {
                "dataset_A": 0.1
                "dataset_B": 0.1
            },
            "test_split": {
                "dataset_A": 0.1
                "dataset_B": 0.1
            }
            # (note that the dataset names above are defined in the field below)
        },
        "datasets": {
            "dataset_A": {
                # type of dataset interface to instantiate
                "type": "...",
                "params": {
                    # ...
                }
            },
            "dataset_B": {
                # type of dataset interface to instantiate
                "type": "...",
                "params": {
                    # ...
                },
                # if it does not derive from 'thelper.data.parsers.Dataset', a task is needed:
                "task": {
                    # this type must derive from 'thelper.tasks.Task'
                    "type": "...",
                    "params": {
                        # ...
                    }
                }
            },
            # ...
        },
        # ...

    Args:
        config: a dictionary that provides all required data configuration information under two fields,
            namely 'datasets' and 'loaders'.
        save_dir: the path to the root directory where the session directory should be saved. Note that
            this is not the path to the session directory itself, but its parent, which may also contain
            other session directories.

    Returns:
        A 4-element tuple that contains: 1) the global task object to specialize models and trainers with;
        2) the training data loader; 3) the validation data loader; and 4) the test data loader.

    .. seealso::
        | :func:`thelper.data.utils.create_parsers`
        | :func:`thelper.transforms.utils.load_augments`
        | :func:`thelper.transforms.utils.load_transforms`
    """
    logstamp = thelper.utils.get_log_stamp()
    repover = thelper.__version__ + ":" + thelper.utils.get_git_stamp()
    session_name = config["name"] if "name" in config else "session"
    data_logger_dir = None
    if save_dir is not None:
        data_logger_dir = os.path.join(save_dir, "logs")
        os.makedirs(data_logger_dir, exist_ok=True)
        data_logger_path = os.path.join(data_logger_dir, "data.log")
        data_logger_format = logging.Formatter("[%(asctime)s - %(process)s] %(levelname)s : %(message)s")
        data_logger_fh = logging.FileHandler(data_logger_path)
        data_logger_fh.setFormatter(data_logger_format)
        logger.addHandler(data_logger_fh)
        logger.info("created data log for session '%s'" % session_name)
        config_backup_path = os.path.join(data_logger_dir, "config." + logstamp + ".json")
        with open(config_backup_path, "w") as fd:
            json.dump(config, fd, indent=4, sort_keys=False)
    logger.debug("loading data usage config")
    # todo: 'data_config' field is deprecated, might be removed later
    if "data_config" in config:
        logger.warning("using 'data_config' field in configuration dictionary is deprecated; switch it to 'loaders'")
    loaders_config = thelper.utils.get_key(["data_config", "loaders"], config)
    # noinspection PyProtectedMember
    from thelper.data.loaders import _LoaderFactory as LoaderFactory
    loader_factory = LoaderFactory(loaders_config)
    datasets, task = create_parsers(config, loader_factory.get_base_transforms())
    if not datasets or task is None:
        raise AssertionError("invalid dataset configuration (got empty list)")
    for dataset_name, dataset in datasets.items():
        logger.info("parsed dataset: %s" % str(dataset))
    logger.info("task info: %s" % str(task))
    logger.debug("splitting datasets and creating loaders...")
    train_idxs, valid_idxs, test_idxs = loader_factory.get_split(datasets, task)
    if save_dir is not None:
        with open(os.path.join(data_logger_dir, "task.log"), "a+") as fd:
            fd.write("session: %s-%s\n" % (session_name, logstamp))
            fd.write("version: %s\n" % repover)
            fd.write(str(task) + "\n")
        for dataset_name, dataset in datasets.items():
            dataset_log_file = os.path.join(data_logger_dir, dataset_name + ".log")
            if not loader_factory.skip_verif and os.path.isfile(dataset_log_file):
                logger.info("verifying sample list for dataset '%s'..." % dataset_name)
                with open(dataset_log_file, "r") as fd:
                    log_content = fd.read()
                    if not log_content or log_content[0] != "{":
                        # could not find new style (json) dataset log, cannot easily parse and compare this log
                        logger.warning("cannot verify that old split is similar to new split, log is out-of-date")
                        continue
                    log_content = json.loads(log_content)
                    if "samples" not in log_content or not isinstance(log_content["samples"], list):
                        raise AssertionError("unexpected dataset log content (bad 'samples' field)")
                    samples_old = log_content["samples"]
                    samples_new = dataset.samples if hasattr(dataset, "samples") and isinstance(dataset.samples, list) else []
                    if len(samples_old) != len(samples_new):
                        query_msg = "Old sample list for dataset '%s' mismatch with current sample list; proceed anyway?"
                        answer = thelper.utils.query_yes_no(query_msg, bypass="n")
                        if not answer:
                            logger.error("sample list mismatch with previous run; user aborted")
                            sys.exit(1)
                        break
                    else:
                        breaking = False
                        for set_name, idxs in zip(["train_idxs", "valid_idxs", "test_idxs"],
                                                  [train_idxs[dataset_name], valid_idxs[dataset_name], test_idxs[dataset_name]]):
                            # index values were paired in tuples earlier, 0=idx, 1=label
                            if log_content[set_name] != [idx for idx, _ in idxs]:
                                query_msg = "Old indices list for dataset '%s' mismatch with current indices" \
                                            "list ('%s'); proceed anyway?" % (dataset_name, set_name)
                                answer = thelper.utils.query_yes_no(query_msg, bypass="n")
                                if not answer:
                                    logger.error("indices list mismatch with previous run; user aborted")
                                    sys.exit(1)
                                breaking = True
                                break
                        if not breaking:
                            for idx, (sample_new, sample_old) in enumerate(zip(samples_new, samples_old)):
                                if str(sample_new) != sample_old:
                                    query_msg = "Old sample #%d for dataset '%s' mismatch with current #%d; proceed anyway?" \
                                                "\n\told: %s\n\tnew: %s" % (idx, dataset_name, idx, str(sample_old), str(sample_new))
                                    answer = thelper.utils.query_yes_no(query_msg, bypass="n")
                                    if not answer:
                                        logger.error("sample list mismatch with previous run; user aborted")
                                        sys.exit(1)
                                    break
        for dataset_name, dataset in datasets.items():
            dataset_log_file = os.path.join(data_logger_dir, dataset_name + ".log")
            samples = dataset.samples if hasattr(dataset, "samples") and isinstance(dataset.samples, list) else []
            log_content = {
                "metadata": {
                    "session_name": session_name,
                    "logstamp": logstamp,
                    "version": repover,
                    "dataset": str(dataset),
                },
                "samples": [str(sample) for sample in samples],
                # index values were paired in tuples earlier, 0=idx, 1=label
                "train_idxs": [idx for idx, _ in train_idxs[dataset_name]],
                "valid_idxs": [idx for idx, _ in valid_idxs[dataset_name]],
                "test_idxs": [idx for idx, _ in test_idxs[dataset_name]]
            }
            # now, always overwrite, as it can get too big otherwise
            with open(dataset_log_file, "w") as fd:
                json.dump(log_content, fd, indent=4, sort_keys=False)
    train_loader, valid_loader, test_loader = loader_factory.create_loaders(datasets, train_idxs, valid_idxs, test_idxs)
    return task, train_loader, valid_loader, test_loader


def create_parsers(config, base_transforms=None):
    """Instantiates dataset parsers based on a provided dictionary.

    This function will instantiate dataset parsers as defined in a name-type-param dictionary. If multiple
    datasets are instantiated, this function will also verify their task compatibility and return the global
    task. The dataset interfaces themselves should be derived from :class:`thelper.data.parsers.Dataset`, be
    compatible with :class:`thelper.data.parsers.ExternalDataset`, or should provide a 'task' field specifying
    all the information related to sample dictionary keys and model i/o. An example configuration is provided
    in :func:`thelper.data.utils.create_parsers`.

    The provided configuration will be parsed for a 'datasets' dictionary entry. The keys in this dictionary
    are treated as unique dataset names and are used for lookups. The value associated to each key (or dataset
    name) should be a type-params dictionary that can be parsed to instantiate the dataset interface.

    An example configuration dictionary is given in :func:`thelper.data.utils.create_loaders`.

    Args:
        config: a dictionary that provides unique dataset names and parameters needed for instantiation under
            the 'datasets' field.
        base_transforms: the transform operation that should be applied to all loaded samples, and that
            will be provided to the constructor of all instantiated dataset parsers.

    Returns:
        A 2-element tuple that contains: 1) the list of dataset interfaces/parsers that were instantiated; and
        2) a task object compatible with all of those (see :class:`thelper.tasks.utils.Task` for more information).

    .. seealso::
        | :func:`thelper.data.utils.create_loaders`
        | :class:`thelper.data.parsers.Dataset`
        | :class:`thelper.data.parsers.ExternalDataset`
        | :class:`thelper.tasks.utils.Task`
    """
    if not isinstance(config, dict):
        raise AssertionError("unexpected session config type")
    if "datasets" not in config or not config["datasets"]:
        raise AssertionError("config missing 'datasets' field (must contain dict or str value)")
    config = config["datasets"]  # no need to keep the full config here
    if isinstance(config, str):
        if os.path.isfile(config) and os.path.splitext(config)[1] == ".json":
            config = json.load(open(config))
        else:
            raise AssertionError("'datasets' string should point to valid json file")
    logger.debug("loading datasets templates")
    if not isinstance(config, dict):
        raise AssertionError("invalid datasets config type (must be dictionary)")
    datasets = {}
    tasks = []
    for dataset_name, dataset_config in config.items():
        logger.debug("loading dataset '%s' configuration..." % dataset_name)
        if "type" not in dataset_config:
            raise AssertionError("missing field 'type' for instantiation of dataset '%s'" % dataset_name)
        dataset_type = thelper.utils.import_class(dataset_config["type"])
        dataset_params = thelper.utils.get_key_def(["params", "parameters"], dataset_config, {})
        transforms = None
        if "transforms" in dataset_config and dataset_config["transforms"]:
            logger.debug("loading custom transforms for dataset '%s'..." % dataset_name)
            transforms = thelper.transforms.load_transforms(dataset_config["transforms"])
            if base_transforms is not None:
                transforms = thelper.transforms.Compose([transforms, base_transforms])
        elif base_transforms is not None:
            transforms = base_transforms
        if issubclass(dataset_type, thelper.data.Dataset):
            # assume that the dataset is derived from thelper.data.parsers.Dataset (it is fully sampling-ready)
            dataset = dataset_type(config=dataset_params, transforms=transforms)
            if "task" in dataset_config:
                logger.warning("'task' field detected in dataset '%s' config; dataset's default task will be ignored" % dataset_name)
                task = thelper.tasks.create_task(dataset_config["task"])
            else:
                task = dataset.get_task()
        else:
            if "task" not in dataset_config or not dataset_config["task"]:
                raise AssertionError("external dataset '%s' must define task interface in its configuration dict" % dataset_name)
            task = thelper.tasks.create_task(dataset_config["task"])
            # assume that __getitem__ and __len__ are implemented, but we need to make it sampling-ready
            dataset = thelper.data.ExternalDataset(dataset_type, task, config=dataset_params, transforms=transforms)
        if task is None:
            raise AssertionError("parsed task interface should not be None anymore (old code doing something strange?)")
        tasks.append(task)
        datasets[dataset_name] = dataset
    return datasets, thelper.tasks.create_global_task(tasks)


def get_class_weights(label_map, stype, invmax, maxw=float('inf'), minw=0.0, norm=True):
    """Returns a map of adjusted class weights based on a given rebalancing strategy.

    Args:
        label_map: map of index lists or sample counts tied to class labels.
        stype: weighting strategy ('uniform', 'linear', or 'rootX'); see :class:`thelper.data.samplers.WeightedSubsetRandomSampler`
            for more information on these.
        invmax: specifies whether to max-invert the weight vector (thus creating cost factors) or not (default=True).
        maxw: maximum allowed weight value (applied after invmax, if required).
        minw: minimum allowed weight value (applied after invmax, if required).
        norm: specifies whether the returned weights should be normalized (default=True, i.e. normalized).

    Returns:
        Map of adjusted weights tied to class labels.

    .. seealso::
        | :class:`thelper.data.samplers.WeightedSubsetRandomSampler`
    """
    if not isinstance(label_map, dict) or any([not isinstance(val, (list, int)) for val in label_map.values()]):
        raise AssertionError("unexpected label map type")
    if stype == "uniform":
        label_weights = {label: 1.0 / len(label_map) for label in label_map}
    elif stype == "linear" or "root" in stype:
        if stype == "root" or stype == "linear":
            rpow = 1.0
        else:
            rpow = 1.0 / float(stype.split("root", 1)[1])
        label_sizes = {label: len(v) if isinstance(v, list) else v for label, v in label_map.items()}
        label_weights = {label: (lsize / sum(label_sizes.values())) ** rpow for label, lsize in label_sizes.items()}
    else:
        raise AssertionError("unknown label weighting strategy")
    if invmax:
        label_weights = {label: max(label_weights.values()) / max(weight, 1e-6) for label, weight in label_weights.items()}
    label_weights = {label: min(max(weight, minw), maxw) for label, weight in label_weights.items()}
    if norm:
        tot_weight = sum([w for w in label_weights.values()])
        label_weights = {label: weight / tot_weight for label, weight in label_weights.items()}
    return label_weights



"""Neural network and model package.

This package contains classes that define blocks and modules used in various neural network
architectures. Most of these classes have been adapted from external sources; see their individual
headers for more information.
"""

import logging

import thelper.nn.utils  # noqa: F401
from thelper.nn.utils import Module  # noqa: F401
from thelper.nn.utils import create_model  # noqa: F401

logger = logging.getLogger("thelper.nn")



"""Transformations operations module.

All transforms should aim to be compatible with both numpy arrays and
PyTorch tensors. By default, images are processed using ``__call__``,
meaning that for a given transformation ``t``, we apply it via::

    image_transformed = t(image)

All important parameters for an operation should also be passed in the
constructor and exposed in the operation's ``__repr__`` function so that
external parsers can discover exactly how to reproduce their behavior.
"""

import bisect
import copy
import itertools
import logging
import math

import cv2 as cv
import numpy as np
import PIL.Image
import torch
import torchvision.transforms.functional
import torchvision.utils

import thelper.utils

logger = logging.getLogger(__name__)


class NoTransformWrapper(object):
    """Used to flag some ops that should not be externally wrapped for sample/key handling."""
    pass


class Compose(torchvision.transforms.Compose):
    """Composes several transforms together (with support for invert ops).

    This interface is fully compatible with ``torchvision.transforms.Compose``.
    """

    def __init__(self, transforms):
        """Forwards the list of transformations to the base class."""
        if not isinstance(transforms, list) or not transforms:
            raise AssertionError("expected transforms to be provided as a non-empty list")
        if all([isinstance(stage, dict) for stage in transforms]):
            transforms, _ = thelper.transforms.load_transforms(transforms)
        super().__init__(transforms)

    def invert(self, sample):
        """Tries to invert the transformations applied to a sample.

        Will throw if one of the transformations cannot be inverted.
        """
        for t in reversed(self.transforms):
            if not hasattr(t, "invert"):
                raise AssertionError("missing invert op for transform = %s" % repr(t))
            sample = t.invert(sample)
        return sample

    def __repr__(self):
        """Provides print-friendly output for class attributes."""
        return self.__class__.__name__ + ": [\n\t" + ",\n\t".join([str(t) for t in self.transforms]) + "]"

    def set_seed(self, seed):
        """Sets the internal seed to use for stochastic ops."""
        if self.transforms is not None:
            for t in self.transforms:
                if hasattr(t, "set_seed") and callable(t.set_seed):
                    t.set_seed(seed)

    def set_epoch(self, epoch=0):
        """Sets the current epoch number in order to change the behavior of some suboperations."""
        if not isinstance(epoch, int) or epoch < 0:
            raise AssertionError("invalid epoch value")
        if self.transforms is not None:
            for t in self.transforms:
                if hasattr(t, "set_epoch") and callable(t.set_epoch):
                    t.set_epoch(epoch)


class CustomStepCompose(torchvision.transforms.Compose):
    """Composes several transforms together based on an epoch schedule.

    This interface is fully compatible with ``torchvision.transforms.Compose``. It can be useful if
    some operations should change their behavior over the course of a training session. Note that
    all epoch indices are assumed to be 0-based.

    Usage example in Python::

        # We will scale the resolution of input patches based on an arbitrary schedule
        # dsize = (16, 16)   if epoch < 2         (16x16 patches before epoch 2)
        # dsize = (32, 32)   if 2 <= epoch < 4    (32x32 patches before epoch 4)
        # dsize = (64, 64)   if 4 <= epoch < 8    (64x64 patches before epoch 8)
        # dsize = (112, 112) if 8 <= epoch < 12   (112x112 patches before epoch 12)
        # dsize = (160, 160) if 12 <= epoch < 15  (160x160 patches before epoch 15)
        # dsize = (196, 196) if 15 <= epoch < 18  (196x196 patches before epoch 18)
        # dsize = (224, 224) if epoch >= 18       (224x224 patches past epoch 18)
        transforms = CustomStepCompose(milestones={
            0: thelper.transforms.Resize(dsize=(16, 16)),
            2: thelper.transforms.Resize(dsize=(32, 32)),
            4: thelper.transforms.Resize(dsize=(64, 64)),
            8: thelper.transforms.Resize(dsize=(112, 112)),
            12: thelper.transforms.Resize(dsize=(160, 160)),
            15: thelper.transforms.Resize(dsize=(196, 196)),
            18: thelper.transforms.Resize(dsize=(224, 224)),
        })
        for epoch in range(100):
            transforms.set_epoch(epoch)
            for sample in loader:
                sample = transforms(sample)
                train(...)

    Attributes:
        stages: list of epochs where a new scaling factor is to be applied.
        transforms: list of transformation to apply at each stage.
        milestones: original milestones map provided in the constructor.
        epoch: index of the current epoch.

    .. seealso::
        | :class:`thelper.transforms.operations.Compose`
    """

    def __init__(self, milestones, last_epoch=-1):
        """Receives the milestone stages (or stage lists), and the initialization state.

        If the milestones do not include the first epoch (idx = 0), then no transform will be applied
        until the next specified epoch index. When last_epoch is -1, the training is assumed to start
        from scratch.

        Args:
            milestones: Map of epoch indices tied to transformation stages. Keys must be increasing.
            last_epoch: The index of last epoch. Default: -1.
        """
        if not isinstance(milestones, dict):
            raise AssertionError("milestones should be provided as a dictionary")
        self.transforms, self.stages = [], []
        if len(milestones) > 0:
            if isinstance(list(milestones.keys())[0], str):  # fixup for json-based config loaders
                self.stages = [int(key) for key in milestones.keys()]
            elif isinstance(list(milestones.keys())[0], int):
                self.stages = list(milestones.keys())
            else:
                raise AssertionError("milestone stages should be epoch indices (integers)")
            if self.stages != sorted(self.stages):
                raise AssertionError("milestone stages should be increasing integers")
        for transforms in milestones.values():
            if isinstance(transforms, list):
                if all([isinstance(t, dict) for t in transforms]):
                    transforms = thelper.transforms.load_transforms(transforms)
                else:
                    if not all([hasattr(t, "__call__") for t in transforms]):
                        raise AssertionError("stage transformations should be callable")
            else:
                if not hasattr(transforms, "__call__"):
                    raise AssertionError("stage transformations should be callable")
            self.transforms.append(transforms)
        if 0 not in self.stages:
            print("prestages = %s" % str(self.stages))
            self.stages.insert(0, int(0))
            print("poststages = %s" % str(self.stages))
            print("pretransforms = %s" % str(self.transforms))
            self.transforms.insert(0, [])
            print("posttransforms = %s" % str(self.transforms))
        self.milestones = milestones
        super().__init__(self.transforms)
        self.epoch = last_epoch + 1

    def _get_stage_idx(self, epoch):
        if epoch in self.stages:
            return self.stages.index(epoch)
        return max(bisect.bisect_right(self.stages, epoch) - 1, 0)

    def invert(self, sample):
        """Tries to invert the transformations applied to a sample.

        Will throw if one of the transformations cannot be inverted.
        """
        transforms = self.transforms[self._get_stage_idx(self.epoch)]
        if isinstance(transforms, list):
            for t in reversed(transforms):
                if not hasattr(t, "invert"):
                    raise AssertionError("missing invert op for transform = %s" % repr(t))
                sample = t.invert(sample)
        else:
            if not hasattr(transforms, "invert"):
                raise AssertionError("missing invert op for transform = %s" % repr(transforms))
            sample = transforms.invert(sample)
        return sample

    def __call__(self, img):
        """Applies the current stage of transformation operations to a sample."""
        print("self.stages = %s" % str(self.stages))
        print("self.epoch = %d" % self.epoch)
        print("len(self.stages) - 1 = %d" % (len(self.stages) - 1))
        print("stage_idx = %d" % self._get_stage_idx(self.epoch))
        transforms = self.transforms[self._get_stage_idx(self.epoch)]
        if isinstance(transforms, list):
            for t in transforms:
                img = t(img)
        else:
            img = transforms(img)
        return img

    def __repr__(self):
        """Provides print-friendly output for class attributes."""
        return self.__class__.__name__ + \
            ": {\n\t" + ",\n\t".join(["%s: %s" % (str(k), str(v)) for k, v in self.milestones.items()]) + "}"

    def set_seed(self, seed):
        """Sets the internal seed to use for stochastic ops."""
        transforms = self.transforms[self._get_stage_idx(self.epoch)]
        if isinstance(transforms, list):
            for t in transforms:
                if hasattr(t, "set_seed") and callable(t.set_seed):
                    t.set_seed(seed)
        else:
            if hasattr(transforms, "set_seed") and callable(transforms.set_seed):
                transforms.set_seed(seed)

    def set_epoch(self, epoch=0):
        """Sets the current epoch number in order to change the behavior of some suboperations."""
        if not isinstance(epoch, int) or epoch < 0:
            raise AssertionError("invalid epoch value")
        transforms = self.transforms[self._get_stage_idx(self.epoch)]
        if isinstance(transforms, list):
            for t in transforms:
                if hasattr(t, "set_epoch") and callable(t.set_epoch):
                    t.set_epoch(epoch)
        else:
            if hasattr(transforms, "set_epoch") and callable(transforms.set_epoch):
                transforms.set_epoch(epoch)
        self.epoch = epoch

    def step(self, epoch=None):  # used for interface compatibility with LRSchedulers only
        """Advances the epoch tracker in order to change the behavior of some suboperations."""
        if epoch is None:
            epoch = self.epoch + 1
        self.epoch = epoch


class ToNumpy(object):
    """Converts and returns an image in numpy format from a ``torch.Tensor`` or ``PIL.Image`` format.

    This operation is deterministic. The returns image will always be encoded as HxWxC, where
    if the input has three channels, the ordering might be optionally changed.

    Attributes:
        reorder_BGR: specifies whether the channels should be reordered in OpenCV format.
    """

    def __init__(self, reorder_BGR=False):
        """Initializes transformation parameters."""
        self.reorder_BGR = reorder_BGR

    def __call__(self, sample):
        """Converts and returns an image in numpy format.

        Args:
            sample: the image to convert; should be a tensor, numpy array, or PIL image.

        Returns:
            The numpy-converted image.
        """
        if isinstance(sample, np.ndarray):
            pass  # no transform needed, channel reordering done at end
        elif isinstance(sample, torch.Tensor):
            sample = np.transpose(sample.cpu().numpy(), [1, 2, 0])  # CxHxW to HxWxC
        elif isinstance(sample, PIL.Image.Image):
            sample = np.asarray(sample)
        else:
            raise AssertionError("unknown image type, cannot process sample")
        if self.reorder_BGR:
            return sample[..., ::-1]  # assumes channels already in last dim
        else:
            return sample

    def invert(self, sample):
        """Specifies that this operation cannot be inverted, as the original data type is unknown."""
        raise AssertionError("cannot be inverted")


class CenterCrop(object):
    """Returns a center crop from a given image via OpenCV and numpy.

    This operation is deterministic. The code relies on OpenCV, meaning the border arguments
    must be compatible with ``cv2.copyMakeBorder``.

    Attributes:
        size: the size of the target crop (tuple of width, height).
        relative: specifies whether the target crop size is relative to the image size or not.
        bordertype: argument forwarded to ``cv2.copyMakeBorder``.
        borderval: argument forwarded to ``cv2.copyMakeBorder``.
    """

    def __init__(self, size, bordertype=cv.BORDER_CONSTANT, borderval=0):
        """Validates and initializes center crop parameters.

        Args:
            size: size of the target crop, provided as tuple or list. If integer values are used,
                the size is assumed to be absolute. If floating point values are used, the size is
                assumed to be relative, and will be determined dynamically for each sample. If a
                tuple is used, it is assumed to be (width, height).
            bordertype: border copy type to use when the image is too small for the required crop size.
                See ``cv2.copyMakeBorder`` for more information.
            borderval: border value to use when the image is too small for the required crop size. See
                ``cv2.copyMakeBorder`` for more information.
        """
        if isinstance(size, (tuple, list)):
            if len(size) != 2:
                raise AssertionError("expected center crop dim input as 2-item tuple (width,height)")
            self.size = tuple(size)
        elif isinstance(size, (int, float)):
            self.size = (size, size)
        else:
            raise AssertionError("unexpected center crop dim input type (need tuple or int/float)")
        if self.size[0] <= 0 or self.size[1] <= 0:
            raise AssertionError("crop dimensions must be positive")
        if isinstance(self.size[0], float):
            self.relative = True
        elif isinstance(self.size[0], int):
            self.relative = False
        else:
            raise AssertionError("unexpected center crop dim input type (need tuple or int/float)")
        self.bordertype = thelper.utils.import_class(bordertype) if isinstance(bordertype, str) else bordertype
        self.borderval = borderval

    def __call__(self, sample):
        """Extracts and returns a central crop from the provided image.

        Args:
            sample: the image to generate the crop from; should be a 2d or 3d numpy array.

        Returns:
            The center crop.
        """
        if not isinstance(sample, np.ndarray):
            raise AssertionError("sample type should be np.ndarray")
        if sample.ndim < 2 or sample.ndim > 3:
            raise AssertionError("bad input dimensions; must be 2-d, or 3-d (with channels)")
        crop_height = int(round(self.size[1] * sample.shape[0])) if self.relative else self.size[1]
        crop_width = int(round(self.size[0] * sample.shape[1])) if self.relative else self.size[0]
        tl = [sample.shape[1] // 2 - crop_width // 2, sample.shape[0] // 2 - crop_height // 2]
        br = [tl[0] + crop_width, tl[1] + crop_height]
        return thelper.utils.safe_crop(sample, tl, br, self.bordertype, self.borderval)

    def invert(self, sample):
        """Specifies that this operation cannot be inverted, as data loss is incurred during image transformation."""
        raise AssertionError("cannot be inverted")

    def __repr__(self):
        """Provides print-friendly output for class attributes."""
        return self.__class__.__name__ + ": {{size: {0}, bordertype: {1}, bordervalue: {2}}}".format(
            self.size, self.bordertype, self.borderval)


class RandomResizedCrop(object):
    """Returns a resized crop of a randomly selected image region.

    This operation is stochastic, and thus cannot be inverted. Each time the operation is called,
    a random check will determine whether a transformation is applied or not. The code relies on
    OpenCV, meaning the interpolation arguments must be compatible with ``cv2.resize``.

    Attributes:
        output_size: size of the output crop, provided as a single element (``edge_size``) or as a
            two-element tuple or list (``[width, height]``). If integer values are used, the size is
            assumed to be absolute. If floating point values are used (i.e. in [0,1]), the output
            size is assumed to be relative to the original image size, and will be determined at
            execution time for each sample. If set to ``None``, the crop will not be resized.
        input_size: range of the input region sizes, provided as a pair of elements
            (``[min_edge_size, max_edge_size]``) or as a pair of tuples or lists
            (``[[min_width, min_height], [max_width, max_height]]``). If the pair-of-pairs format is
            used, the ``ratio`` argument cannot be used. If integer values are used, the ranges are
            assumed to be absolute. If floating point values are used (i.e. in [0,1]), the ranges
            are assumed to be relative to the original image size, and will be determined at
            execution time for each sample.
        ratio: range of minimum/maximum input region aspect ratios to use. This argument cannot be
            used if the pair-of-pairs format is used for the ``input_size`` argument.
        probability: the probability that the transformation will be applied when called; if not
            applied, the returned image will be the original.
        random_attempts: the number of random sampling attempts to try before reverting to center
            or most-probably-valid crop generation.
        min_roi_iou: minimum roi intersection over union (IoU) required for accepting a tile (in [0,1]).
        flags: interpolation flag forwarded to ``cv2.resize``.
    """

    def __init__(self, output_size, input_size=(0.08, 1.0), ratio=(0.75, 1.33), probability=1.0,
                 random_attempts=10, min_roi_iou=1.0, flags=cv.INTER_LINEAR):
        """Validates and initializes center crop parameters.

        Args:
            output_size: size of the output crop, provided as a single element (``edge_size``) or as a
                two-element tuple or list (``[width, height]``). If integer values are used, the size is
                assumed to be absolute. If floating point values are used (i.e. in [0,1]), the output
                size is assumed to be relative to the original image size, and will be determined at
                execution time for each sample. If set to ``None``, the crop will not be resized.
            input_size: range of the input region sizes, provided as a pair of elements
                (``[min_edge_size, max_edge_size]``) or as a pair of tuples or lists
                (``[[min_width, min_height], [max_width, max_height]]``). If the pair-of-pairs format is
                used, the ``ratio`` argument cannot be used. If integer values are used, the ranges are
                assumed to be absolute. If floating point values are used (i.e. in [0,1]), the ranges
                are assumed to be relative to the original image size, and will be determined at
                execution time for each sample.
            ratio: range of minimum/maximum input region aspect ratios to use. This argument cannot be
                used if the pair-of-pairs format is used for the ``input_size`` argument.
            probability: the probability that the transformation will be applied when called; if not
                applied, the returned image will be the original.
            random_attempts: the number of random sampling attempts to try before reverting to center
                or most-probably-valid crop generation.
            min_roi_iou: minimum roi intersection over union (IoU) required for producing a tile.
            flags: interpolation flag forwarded to ``cv2.resize``.
        """
        if isinstance(output_size, (tuple, list)):
            if len(output_size) != 2:
                raise AssertionError("expected output size to be two-element list or tuple, or single scalar")
            if not all([isinstance(s, int) for s in output_size]) and not all([isinstance(s, float) for s in output_size]):
                raise AssertionError("expected output size pair elements to be the same type (int or float)")
            self.output_size = output_size
        elif isinstance(output_size, (int, float)):
            self.output_size = (output_size, output_size)
        elif output_size is None:
            self.output_size = None
        else:
            raise AssertionError("unexpected output size type (need tuple/list/int/float)")
        if self.output_size is not None:
            for s in self.output_size:
                if not ((isinstance(s, float) and 0 < s <= 1) or (isinstance(s, int) and s > 0)):
                    raise AssertionError("invalid output size value (%s)" % str(s))
        if not isinstance(input_size, (tuple, list)) or len(input_size) != 2:
            raise AssertionError("expected input size to be provided as a pair of elements or a pair of tuples/lists")
        if all([isinstance(s, int) for s in input_size]) or all([isinstance(s, float) for s in input_size]):
            if ratio is not None and isinstance(ratio, (tuple, list)):
                if len(ratio) != 2:
                    raise AssertionError("invalid ratio tuple/list length (expected two elements)")
                if not all([isinstance(r, float) for r in ratio]):
                    raise AssertionError("expected ratio pair elements to be float values")
                self.ratio = (min(ratio), max(ratio))
            elif ratio is not None and isinstance(ratio, float):
                self.ratio = (ratio, ratio)
            else:
                raise AssertionError("invalid aspect ratio, expected 2-element tuple/list or single float")
            self.input_size = ((min(input_size), min(input_size)), (max(input_size), max(input_size)))
        elif all([isinstance(s, tuple) and len(s) == 2 for s in input_size]) or \
                all([isinstance(s, list) and len(s) == 2 for s in input_size]):
            if ratio is not None and isinstance(ratio, (tuple, list)) and len(ratio) != 0:
                raise AssertionError("cannot specify input sizes in two-element tuples/lists and also provide aspect ratios")
            for t in input_size:
                for s in t:
                    if not isinstance(s, (int, float)) or isinstance(type(s), type(input_size[0][0])):
                        raise AssertionError("input sizes should all be same type, either int or float")
            self.input_size = (min(input_size[0][0], input_size[1][0]), min(input_size[0][1], input_size[1][1]),
                               max(input_size[0][0], input_size[1][0]), max(input_size[0][1], input_size[1][1]))
            self.ratio = None  # ignored since input_size contains all necessary info
        else:
            raise AssertionError("expected input size to be two-elem list/tuple of int/float or two-element list/tuple of int/float")
        for t in self.input_size:
            for s in t:
                if not ((isinstance(s, float) and 0 < s <= 1) or (isinstance(s, int) and s > 0)):
                    raise AssertionError("invalid input size value (%s)" % str(s))
        if probability < 0 or probability > 1:
            raise AssertionError("invalid probability value (should be in [0,1]")
        self.probability = probability
        if random_attempts <= 0:
            raise AssertionError("invalid random_attempts value (should be > 0)")
        self.random_attempts = random_attempts
        if not isinstance(min_roi_iou, float) or min_roi_iou < 0 or min_roi_iou > 1:
            raise AssertionError("invalid minimum roi IoU score (should be float in [0,1])")
        self.min_roi_iou = min_roi_iou
        self.flags = thelper.utils.import_class(flags) if isinstance(flags, str) else flags
        self.warned_no_crop_found_with_roi = False

    def __call__(self, image, roi=None, mask=None, bboxes=None):
        """Extracts and returns a random (resized) crop from the provided image.

        Args:
            image: the image to generate the crop from. If given as a 2-element list, it is assumed to
                contain both the image and the roi (passed through a composer).
            roi: the roi to check tile intersections with (may be ``None``).
            mask: a mask to crop simultaneously with the input image (may be ``None``).
            bboxes: a list or array of bounding boxes (in xywh format) to crop with the input image
                (may be ``None``).

        Returns:
            The randomly selected and resized crop. If mask and/or bboxes is given, the output will be
            a dictionary containing the results under the ``image``, ``mask``, and ``bboxes`` keys.
        """
        if isinstance(image, list) and len(image) == 2:
            if roi is not None:
                raise AssertionError("roi provided twice")
            # we assume that the roi was given as the 2nd element of the list
            image, roi = image[0], image[1]
        if isinstance(image, PIL.Image.Image):
            image = np.asarray(image)
        elif not isinstance(image, np.ndarray):
            raise AssertionError("image type should be np.ndarray")
        if self.probability < 1 and np.random.uniform(0, 1) > self.probability:
            return image
        image_height, image_width = image.shape[0], image.shape[1]
        target_height, target_width = None, None
        target_row, target_col = None, None
        for attempt in range(self.random_attempts):
            if self.ratio is None:
                target_width = np.random.uniform(self.input_size[0][0], self.input_size[1][0])
                target_height = np.random.uniform(self.input_size[0][1], self.input_size[1][1])
                if isinstance(self.input_size[0][0], float):
                    target_width *= image_width
                    target_height *= image_height
                target_width = int(round(target_width))
                target_height = int(round(target_height))
            elif isinstance(self.input_size[0][0], (int, float)):
                if isinstance(self.input_size[0][0], float):
                    area = image_height * image_width
                    target_area = np.random.uniform(self.input_size[0][0], self.input_size[1][0]) * area
                else:
                    target_area = np.random.uniform(self.input_size[0][0], self.input_size[1][0]) ** 2
                aspect_ratio = np.random.uniform(*self.ratio)
                target_width = int(round(math.sqrt(target_area * aspect_ratio)))
                target_height = int(round(math.sqrt(target_area / aspect_ratio)))
                if np.random.random() < 0.5:
                    target_width, target_height = target_height, target_width
            else:
                raise AssertionError("unhandled crop strategy")
            if target_width <= image_width and target_height <= image_height:
                target_col = np.random.randint(min(0, image_width - target_width), max(0, image_width - target_width) + 1)
                target_row = np.random.randint(min(0, image_height - target_height), max(0, image_height - target_height) + 1)
                if roi is None:
                    break
                roi = thelper.utils.safe_crop(roi, (target_col, target_row), (target_col + target_width, target_row + target_height))
                if np.count_nonzero(roi) >= target_width * target_height * self.min_roi_iou:
                    break
        if target_row is None or target_col is None:
            # fallback, use centered crop
            target_width = target_height = min(image.shape[0], image.shape[1])
            target_col = (image.shape[1] - target_width) // 2
            target_row = (image.shape[0] - target_height) // 2
            if roi is not None and not self.warned_no_crop_found_with_roi:
                logger.warning("random resized crop failing to find proper ROI matches after max attempt count")
                self.warned_no_crop_found_with_roi = True
        crop = thelper.utils.safe_crop(image, (target_col, target_row), (target_col + target_width, target_row + target_height))
        if self.output_size is None:
            return crop
        elif isinstance(self.output_size[0], float):
            output_width = int(round(self.output_size[0] * image.shape[1]))
            output_height = int(round(self.output_size[1] * image.shape[0]))
            return cv.resize(crop, (output_width, output_height), interpolation=self.flags)
        elif isinstance(self.output_size[0], int):
            return cv.resize(crop, (self.output_size[0], self.output_size[1]), interpolation=self.flags)
        else:
            raise AssertionError("unhandled crop strategy")

    def invert(self, image):
        """Specifies that this operation cannot be inverted, as data loss is incurred during image transformation."""
        raise AssertionError("cannot be inverted")

    def __repr__(self):
        """Provides print-friendly output for class attributes."""
        format_str = self.__class__.__name__ + ": {"
        format_str += "output_size: {}, ".format(self.output_size)
        format_str += "input_size: {}, ".format(self.input_size)
        format_str += "ratio: {}, ".format(self.ratio)
        format_str += "probability: {}, ".format(self.probability)
        format_str += "random_attempts: {}, ".format(self.random_attempts)
        format_str += "flags: {}}}".format(self.flags)
        return format_str

    def set_seed(self, seed):
        """Sets the internal seed to use for stochastic ops."""
        np.random.seed(seed)


class Resize(object):
    """Resizes a given image using OpenCV and numpy.

    This operation is deterministic. The code relies on OpenCV, meaning the interpolation arguments
    must be compatible with ``cv2.resize``.

    Attributes:
        interp: interpolation type to use (forwarded to ``cv2.resize``)
        buffer: specifies whether a destination buffer should be used to avoid allocations
        dsize: target image size (tuple of width, height).
        dst: buffer used to avoid reallocations if ``self.buffer == True``
        fx: argument forwarded to ``cv2.resize``.
        fy: argument forwarded to ``cv2.resize``.
    """

    def __init__(self, dsize, fx=0, fy=0, interp=cv.INTER_LINEAR, buffer=False):
        """Validates and initializes resize parameters.

        Args:
            dsize: size of the target image, forwarded to ``cv2.resize``.
            fx: x-scaling factor, forwarded to ``cv2.resize``.
            fy: y-scaling factor, forwarded to ``cv2.resize``.
            interp: resize interpolation type, forwarded to ``cv2.resize``.
            buffer: specifies whether a destination buffer should be used to avoid allocations
        """
        self.interp = thelper.utils.import_class(interp) if isinstance(interp, str) else interp
        self.buffer = buffer
        if isinstance(dsize, tuple):
            self.dsize = dsize
        else:
            self.dsize = tuple(dsize)
        self.dst = None
        self.fx = fx
        self.fy = fy
        if fx < 0 or fy < 0:
            raise AssertionError("scale factors should be null (ignored) or positive")
        if dsize[0] < 0 or dsize[1] < 0:
            raise AssertionError("destination image size should be null (ignored) or positive")
        if fx == 0 and fy == 0 and dsize[0] == 0 and dsize[1] == 0:
            raise AssertionError("need to specify either destination size or scale factor(s)")

    def __call__(self, sample):
        """Returns a resized copy of the provided image.

        Args:
            sample: the image to resize; should be a 2d or 3d numpy array.

        Returns:
            The resized image. May be allocated on the spot, or be a pointer to a local buffer.
        """
        if isinstance(sample, PIL.Image.Image):
            sample = np.asarray(sample)
        elif not isinstance(sample, np.ndarray):
            raise AssertionError("sample type should be np.ndarray")
        if sample.ndim < 2 or sample.ndim > 3:
            raise AssertionError("bad input dimensions; must be 2-d, or 3-d (with channels)")
        if sample.ndim < 3 or sample.shape[2] <= 4:
            if self.buffer:
                cv.resize(sample, self.dsize, dst=self.dst, fx=self.fx, fy=self.fy, interpolation=self.interp)
                if self.dst.ndim == 2:
                    return np.expand_dims(self.dst, 2)
                return self.dst
            else:
                dst = cv.resize(sample, self.dsize, fx=self.fx, fy=self.fy, interpolation=self.interp)
                if dst.ndim == 2:
                    dst = np.expand_dims(dst, 2)
                return dst
        else:  # too many channels, need to split-resize
            slices_dst = self.dst if self.buffer else None
            slices = np.split(sample, sample.shape[2], 2)
            if slices_dst is None or not isinstance(slices_dst, list) or len(slices_dst) != len(slices):
                slices_dst = [None] * len(slices)
            for idx in range(len(slices)):
                cv.resize(slice[idx], self.dsize, dst=slices_dst[idx], fx=self.fx, fy=self.fy, interpolation=self.interp)
            if self.buffer:
                self.dst = slices_dst
            return np.stack(slices_dst, 2)

    def invert(self, sample):
        """Specifies that this operation cannot be inverted, as data loss is incurred during image transformation."""
        # todo, could implement if original size is fixed & known
        raise AssertionError("missing implementation")

    def __repr__(self):
        """Provides print-friendly output for class attributes."""
        return self.__class__.__name__ + ": {{dsize: {0}, fx: {1}, fy: {2}, interp: {3}}}".format(self.dsize, self.fx,
                                                                                                  self.fy, self.interp)


class Affine(object):
    """Warps a given image using an affine matrix via OpenCV and numpy.

    This operation is deterministic. The code relies on OpenCV, meaning the border arguments
    must be compatible with ``cv2.warpAffine``.

    Attributes:
        transf: the 2x3 transformation matrix passed to ``cv2.warpAffine``.
        out_size: target image size (tuple of width, height). If None, same as original.
        flags: extra warp flags forwarded to ``cv2.warpAffine``.
        border_mode: border extrapolation mode forwarded to ``cv2.warpAffine``.
        border_val: border constant extrapolation value forwarded to ``cv2.warpAffine``.
    """

    def __init__(self, transf, out_size=None, flags=None, border_mode=None, border_val=None):
        """Validates and initializes affine warp parameters.

        Args:
            transf: the 2x3 transformation matrix passed to ``cv2.warpAffine``.
            out_size: target image size (tuple of width, height). If None, same as original.
            flags: extra warp flags forwarded to ``cv2.warpAffine``.
            border_mode: border extrapolation mode forwarded to ``cv2.warpAffine``.
            border_val: border constant extrapolation value forwarded to ``cv2.warpAffine``.
        """
        if isinstance(transf, np.ndarray):
            if transf.size != 6:
                raise AssertionError("transformation matrix must be 2x3")
            self.transf = transf.reshape((2, 3)).astype(np.float32)
        elif isinstance(transf, list):
            if not len(transf) == 6:
                raise AssertionError("transformation matrix must be 6 elements (2x3)")
            self.transf = np.asarray(transf).reshape((2, 3)).astype(np.float32)
        else:
            raise AssertionError("unexpected transformation matrix type")
        self.out_size = None
        if out_size is not None:
            if isinstance(out_size, list):
                if len(out_size) != 2:
                    raise AssertionError("output image size should be 2-elem list or tuple")
                self.out_size = tuple(out_size)
            elif isinstance(out_size, tuple):
                if len(out_size) != 2:
                    raise AssertionError("output image size should be 2-elem list or tuple")
                self.out_size = out_size
            else:
                raise AssertionError("unexpected output size type")
        self.flags = flags
        if self.flags is None:
            self.flags = cv.INTER_LINEAR
        self.border_mode = border_mode
        if self.border_mode is None:
            self.border_mode = cv.BORDER_CONSTANT
        self.border_val = border_val
        if self.border_val is None:
            self.border_val = 0

    def __call__(self, sample):
        """Warps a given image using an affine matrix.

        Args:
            sample: the image to warp; should be a 2d or 3d numpy array.

        Returns:
            The warped image.
        """
        if isinstance(sample, PIL.Image.Image):
            sample = np.asarray(sample)
        elif not isinstance(sample, np.ndarray):
            raise AssertionError("sample type should be np.ndarray")
        out_size = self.out_size
        if out_size is None:
            out_size = (sample.shape[1], sample.shape[0])
        return cv.warpAffine(sample, self.transf, dsize=out_size, flags=self.flags,
                             borderMode=self.border_mode, borderValue=self.border_val)

    def invert(self, sample):
        """Inverts the warp transformation, but only is the output image has not been cropped before."""
        if isinstance(sample, PIL.Image.Image):
            sample = np.asarray(sample)
        out_size = self.out_size
        if out_size is None:
            out_size = (sample.shape[1], sample.shape[0])
        else:
            raise AssertionError("unknown original image size, cannot invert affine transform")
        return cv.warpAffine(sample, self.transf, dsize=out_size, flags=self.flags ^ cv.WARP_INVERSE_MAP,
                             borderMode=self.border_mode, borderValue=self.border_val)

    def __repr__(self):
        """Provides print-friendly output for class attributes."""
        return self.__class__.__name__ + ": {{transf: {0}, out_size: {1}}}".format(np.array2string(self.transf), self.out_size)


class RandomShift(object):
    """Randomly translates an image in a provided range via OpenCV and numpy.

    This operation is stochastic, and thus cannot be inverted. Each time the operation is called,
    a random check will determine whether a transformation is applied or not. The code relies on
    OpenCV, meaning the border arguments must be compatible with ``cv2.warpAffine``.

    Attributes:
        min: the minimum pixel shift that can be applied stochastically.
        max: the maximum pixel shift that can be applied stochastically.
        probability: the probability that the transformation will be applied when called.
        flags: extra warp flags forwarded to ``cv2.warpAffine``.
        border_mode: border extrapolation mode forwarded to ``cv2.warpAffine``.
        border_val: border constant extrapolation value forwarded to ``cv2.warpAffine``.
    """

    def __init__(self, min, max, probability=1.0, flags=None, border_mode=None, border_val=None):
        """Validates and initializes shift parameters.

        Args:
            min: the minimum pixel shift that can be applied stochastically.
            max: the maximum pixel shift that can be applied stochastically.
            probability: the probability that the transformation will be applied when called.
            flags: extra warp flags forwarded to ``cv2.warpAffine``.
            border_mode: border extrapolation mode forwarded to ``cv2.warpAffine``.
            border_val: border constant extrapolation value forwarded to ``cv2.warpAffine``.
        """
        if isinstance(min, tuple) and isinstance(max, tuple):
            if len(min) != len(max) or len(min) != 2:
                raise AssertionError("min/max shift tuple must be 2-elem")
            self.min = min
            self.max = max
        elif isinstance(min, list) and isinstance(max, list):
            if len(min) != len(max) or len(min) != 2:
                raise AssertionError("min/max shift list must be 2-elem")
            self.min = tuple(min)
            self.max = tuple(max)
        elif isinstance(min, (int, float)) and isinstance(max, (int, float)):
            self.min = (min, min)
            self.max = (max, max)
        else:
            raise AssertionError("unexpected min/max combo types")
        if self.max[0] < self.min[0] or self.max[1] < self.min[1]:
            raise AssertionError("bad min/max values")
        if probability < 0 or probability > 1:
            raise AssertionError("bad probability range")
        self.probability = probability
        self.flags = flags
        if self.flags is None:
            self.flags = cv.INTER_LINEAR
        self.border_mode = border_mode
        if self.border_mode is None:
            self.border_mode = cv.BORDER_CONSTANT
        self.border_val = border_val
        if self.border_val is None:
            self.border_val = 0

    def __call__(self, sample):
        """Translates a given image using a predetermined min/max range.

        Args:
            sample: the image to translate; should be a 2d or 3d numpy array.

        Returns:
            The translated image.
        """
        if isinstance(sample, PIL.Image.Image):
            sample = np.asarray(sample)
        elif not isinstance(sample, np.ndarray):
            raise AssertionError("sample type should be np.ndarray")
        if self.probability < 1 and np.random.uniform(0, 1) > self.probability:
            return sample
        out_size = (sample.shape[1], sample.shape[0])
        x_shift = np.random.uniform(self.min[0], self.max[0])
        y_shift = np.random.uniform(self.min[1], self.max[1])
        transf = np.float32([[1, 0, x_shift], [0, 1, y_shift]])
        return cv.warpAffine(sample, transf, dsize=out_size, flags=self.flags, borderMode=self.border_mode, borderValue=self.border_val)

    def invert(self, sample):
        """Specifies that this operation cannot be inverted, as it is stochastic, and data loss occurs during transformation."""
        raise AssertionError("operation cannot be inverted")

    def __repr__(self):
        """Provides print-friendly output for class attributes."""
        return self.__class__.__name__ + ": {{min: {0}, max: {1}, prob: {2}}}".format(self.min, self.max, self.probability)

    def set_seed(self, seed):
        """Sets the internal seed to use for stochastic ops."""
        np.random.seed(seed)


class Transpose(object):
    """Transposes an image via numpy.

    This operation is deterministic.

    Attributes:
        axes: the axes on which to apply the transpose; forwarded to ``numpy.transpose``.
        axes_inv: used to invert the tranpose; also forwarded to ``numpy.transpose``.
    """

    def __init__(self, axes):
        """Validates and initializes tranpose parameters.

        Args:
            axes: the axes on which to apply the transpose; forwarded to ``numpy.transpose``.
        """
        axes = np.asarray(axes)
        if axes.ndim > 1:
            raise AssertionError("tranpose param should be 1-d")
        if np.any(axes >= axes.size):
            raise AssertionError("oob dim in axes")
        self.axes = axes
        self.axes_inv = np.asarray([None] * axes.size)
        for i in list(range(len(axes))):
            self.axes_inv[self.axes[i]] = i

    def __call__(self, sample):
        """Transposes a given image.

        Args:
            sample: the image to transpose; should be a numpy array.

        Returns:
            The transposed image.
        """
        if not isinstance(sample, np.ndarray):
            raise AssertionError("sample type should be np.ndarray")
        return np.transpose(sample, self.axes)

    def invert(self, sample):
        """Invert-transposes a given image.

        Args:
            sample: the image to invert-transpose; should be a numpy array.

        Returns:
            The invert-transposed image.
        """
        return np.transpose(sample, self.axes_inv)

    def __repr__(self):
        """Provides print-friendly output for class attributes."""
        return self.__class__.__name__ + ": {{axes: {0}}}".format(self.axes)


class Duplicator(NoTransformWrapper):
    """Duplicates and returns a list of copies of the input sample.

    This operation is used in data augmentation pipelines that rely on probabilistic or preset transformations.
    It can produce a fixed number of simple copies or deep copies of the input samples as required.

    .. warning::
        Since the duplicates will be given directly to the data loader as part of the same minibatch, using too
        many copies can adversely affect gradient descent for that minibatch. To simply increase the total size
        of the training set while still allowing a proper shuffling of samples and/or to keep the minibatch size
        intact, we instead recommend setting the ``train_scale`` configuration value in the data loader. See
        :func:`thelper.data.utils.create_loaders` for more information.

    Attributes:
        count: number of copies to generate.
        deepcopy: specifies whether to deep-copy samples or not.
    """

    def __init__(self, count, deepcopy=False):
        """Validates and initializes duplication parameters.

        Args:
            count: number of copies to generate.
            deepcopy: specifies whether to deep-copy samples or not.
        """
        if count <= 0:
            raise AssertionError("invalid copy count")
        self.count = count
        self.deepcopy = deepcopy

    def __call__(self, sample):
        """Generates and returns duplicates of the sample/object.

        If a dictionary is provided, its values will be expanded into lists that
        will contain all duplicates. Otherwise, the duplicates will be returned
        directly as a list.

        Args:
            sample: the sample/object to duplicate.

        Returns:
            A list of duplicated samples, or a dictionary of duplicate lists.
        """
        copyfct = copy.deepcopy if self.deepcopy else copy.copy
        if isinstance(sample, dict):
            return {k: [copyfct(v) for _ in range(self.count)] for k, v in sample.items()}
        else:
            return [copyfct(sample) for _ in range(self.count)]

    def invert(self, sample):
        """Returns the first instance of the list of duplicates."""
        if isinstance(sample, dict):
            if not all([isinstance(v, list) for v in sample.values()]):
                raise AssertionError("invalid sample type (should be list)")
            if any([len(v) != self.count for v in sample.values()]):
                raise AssertionError("invalid sample list length")
            return {k: v[0] for k, v in sample.items()}
        else:
            if not isinstance(sample, list):
                raise AssertionError("invalid sample type (should be list)")
            if len(sample) != self.count:
                raise AssertionError("invalid sample list length")
            return sample[0]

    def __repr__(self):
        """Provides print-friendly output for class attributes."""
        return self.__class__.__name__ + ": {{count: {0}, deepcopy: {1}}}".format(self.count, self.deepcopy)


class Tile(object):
    """Returns a list of tiles cut out from a given image.

    This operation can perform tiling given an optional mask with a target intersection over union (IoU) score,
    and with an optional overlap between tiles. The tiling is deterministic and can thus be inverted, but only
    if a mask is not used, as some image regions may be lost otherwise.

    If a mask is used, the first tile position is tested exhaustively by iterating over all input coordinates
    starting from the top-left corner of the image. Otherwise, the first tile position is set as (0,0). Then,
    all other tiles are found by offsetting frm these coordinates, and testing for IoU with the mask (if needed).

    Attributes:
        tile_size: size of the output tiles, provided as a single element (``edge_size``) or as a
            two-element tuple or list (``[width, height]``). If integer values are used, the size is
            assumed to be absolute. If floating point values are used (i.e. in [0,1]), the output
            size is assumed to be relative to the original image size, and will be determined at
            execution time for each image.
        tile_overlap: overlap allowed between two neighboring tiles; should be a ratio in [0,1].
        min_mask_iou: minimum mask intersection over union (IoU) required for accepting a tile (in [0,1]).
        offset_overlap: specifies whether the overlap tiling should be offset outside the image or not.
        bordertype: border copy type to use when the image is too small for the required crop size.
            See ``cv2.copyMakeBorder`` for more information.
        borderval: border value to use when the image is too small for the required crop size. See
            ``cv2.copyMakeBorder`` for more information.
    """

    def __init__(self, tile_size, tile_overlap=0.0, min_mask_iou=1.0, offset_overlap=False, bordertype=cv.BORDER_CONSTANT, borderval=0):
        """Validates and initializes tiling parameters.

        Args:
            tile_size: size of the output tiles, provided as a single element (``edge_size``) or as a
                two-element tuple or list (``[width, height]``). If integer values are used, the size is
                assumed to be absolute. If floating point values are used (i.e. in [0,1]), the output
                size is assumed to be relative to the original image size, and will be determined at
                execution time for each image.
            tile_overlap: overlap ratio between two consecutive (neighboring) tiles; should be in [0,1].
            min_mask_iou: minimum mask intersection over union (IoU) required for producing a tile.
            offset_overlap: specifies whether the overlap tiling should be offset outside the image or not.
            bordertype: border copy type to use when the image is too small for the required crop size.
                See ``cv2.copyMakeBorder`` for more information.
            borderval: border value to use when the image is too small for the required crop size. See
                ``cv2.copyMakeBorder`` for more information.
        """
        if isinstance(tile_size, (tuple, list)):
            if len(tile_size) != 2:
                raise AssertionError("expected tile size to be two-element list or tuple, or single scalar")
            if not all([isinstance(s, int) for s in tile_size]) and not all([isinstance(s, float) for s in tile_size]):
                raise AssertionError("expected tile size pair elements to be the same type (int or float)")
            self.tile_size = tile_size
        elif isinstance(tile_size, (int, float)):
            self.tile_size = (tile_size, tile_size)
        else:
            raise AssertionError("unexpected tile size type (need tuple/list/int/float)")
        if not isinstance(tile_overlap, float) or tile_overlap < 0 or tile_overlap >= 1:
            raise AssertionError("invalid tile overlap (should be float in [0,1[)")
        self.tile_overlap = tile_overlap
        if not isinstance(min_mask_iou, float) or min_mask_iou < 0 or min_mask_iou > 1:
            raise AssertionError("invalid minimum mask IoU score (should be float in [0,1])")
        self.min_mask_iou = min_mask_iou
        self.offset_overlap = offset_overlap
        self.bordertype = thelper.utils.import_class(bordertype) if isinstance(bordertype, str) else bordertype
        self.borderval = borderval

    def __call__(self, image, mask=None):
        """Extracts and returns a list of tiles cut out from the given image.

        Args:
            image: the image to cut into tiles. If given as a 2-element list, it is assumed to contain both
                the image and the mask (passed through a composer).
            mask: the mask to check tile intersections with (may be ``None``).

        Returns:
            A list of tiles (numpy-compatible images).
        """
        if isinstance(image, list) and len(image) == 2:
            if mask is not None:
                raise AssertionError("mask provided twice")
            # we assume that the mask was given as the 2nd element of the list
            image, mask = image[0], image[1]
        tile_rects, tile_images = self._get_tile_rects(image, mask), []
        for rect in tile_rects:
            tile_images.append(thelper.utils.safe_crop(image, (rect[0], rect[1]),
                                                       (rect[0]+rect[2], rect[1]+rect[3]),
                                                       self.bordertype, self.borderval))
        return tile_images

    def count_tiles(self, image, mask=None):
        """Returns the number of tiles that would be cut out from the given image.

        Args:
            image: the image to cut into tiles. If given as a 2-element list, it is assumed to contain both
                the image and the mask (passed through a composer).
            mask: the mask to check tile intersections with (may be ``None``).

        Returns:
            The number of tiles that would be cut with :func:`thelper.transforms.operations.Tile.__call__`.
        """
        if isinstance(image, list) and len(image) == 2:
            if mask is not None:
                raise AssertionError("mask provided twice")
            # we assume that the mask was given as the 2nd element of the list
            image, mask = image[0], image[1]
        return len(self._get_tile_rects(image, mask))

    def _get_tile_rects(self, image, mask=None):
        if isinstance(image, PIL.Image.Image):
            image = np.asarray(image)
        elif not isinstance(image, np.ndarray):
            raise AssertionError("image type should be np.ndarray")
        if mask is not None:
            if isinstance(mask, PIL.Image.Image):
                mask = np.asarray(mask)
            elif not isinstance(mask, np.ndarray):
                raise AssertionError("mask type should be np.ndarray")
        tile_rects = []
        height, width = image.shape[0], image.shape[1]
        if isinstance(self.tile_size[0], float):
            tile_size = (int(round(self.tile_size[0] * width)), int(round(self.tile_size[1] * height)))
        else:
            tile_size = self.tile_size
        overlap = (int(round(tile_size[0] * self.tile_overlap)), int(round(tile_size[1] * self.tile_overlap)))
        overlap_offset = (-overlap[0] // 2, -overlap[1] // 2) if self.offset_overlap else (0, 0)
        step_size = (max(tile_size[0] - (overlap[0] // 2) * 2, 1), max(tile_size[1] - (overlap[1] // 2) * 2, 1))
        req_mask_area = tile_size[0] * tile_size[1] * self.min_mask_iou
        if mask is not None:
            if height != mask.shape[0] or width != mask.shape[1]:
                raise AssertionError("image and mask dimensions mismatch")
            if mask.ndim != 2:
                raise AssertionError("mask should be 2d binary (uchar) array")
            offset_coord = None
            row_range = range(overlap_offset[1], height - overlap_offset[1] - tile_size[1] + 1)
            col_range = range(overlap_offset[0], width - overlap_offset[0] - tile_size[0] + 1)
            for row, col in itertools.product(row_range, col_range):
                crop = thelper.utils.safe_crop(mask, (col, row), (col + tile_size[0], row + tile_size[1]))
                if np.count_nonzero(crop) >= req_mask_area:
                    offset_coord = (overlap_offset[0] + ((col - overlap_offset[0]) % step_size[0]),
                                    overlap_offset[1] + ((row - overlap_offset[1]) % step_size[1]))
                    break
            if offset_coord is None:
                return tile_rects
        else:
            offset_coord = overlap_offset
        row = offset_coord[1]
        while row + tile_size[1] <= height - overlap_offset[1]:
            col = offset_coord[0]
            while col + tile_size[0] <= width - overlap_offset[0]:
                if mask is not None:
                    crop = thelper.utils.safe_crop(mask, (col, row), (col + tile_size[0], row + tile_size[1]))
                    if np.count_nonzero(crop) >= req_mask_area:
                        tile_rects.append((col, row, tile_size[0], tile_size[1]))  # rect = (x, y, w, h)
                else:
                    tile_rects.append((col, row, tile_size[0], tile_size[1]))  # rect = (x, y, w, h)
                col += step_size[0]
            row += step_size[1]
        return tile_rects

    def invert(self, image, mask=None):
        """Returns the reconstituted image from a list of tiles, or throws if a mask was used."""
        if mask is not None:
            raise AssertionError("cannot invert operation, mask might have forced the loss of image content")
        else:
            raise NotImplementedError  # TODO

    def __repr__(self):
        """Provides print-friendly output for class attributes."""
        return self.__class__.__name__ + ": {{tile_size: {0}, tile_overlap: {1}, min_mask_iou: {2}}}".format(
            self.tile_size, self.tile_overlap, self.min_mask_iou)


class NormalizeZeroMeanUnitVar(object):
    """Normalizes a given image using a set of mean and standard deviation parameters.

    The samples will be transformed such that ``s = (s - mean) / std``.

    This can be used for whitening; see https://en.wikipedia.org/wiki/Whitening_transformation
    for more information. Note that this operation is also not restricted to images.

    Attributes:
        mean: an array of mean values to subtract from data samples.
        std: an array of standard deviation values to divide with.
        out_type: the output data type to cast the normalization result to.
    """

    def __init__(self, mean, std, out_type=np.float32):
        """Validates and initializes normalization parameters.

        Args:
            mean: an array of mean values to subtract from data samples.
            std: an array of standard deviation values to divide with.
            out_type: the output data type to cast the normalization result to.
        """
        self.out_type = out_type
        self.mean = np.asarray(mean).astype(out_type)
        self.std = np.asarray(std).astype(out_type)
        if self.mean.ndim != 1 or self.std.ndim != 1:
            raise AssertionError("normalization params should be 1-d")
        if self.mean.size != self.std.size:
            raise AssertionError("normalization params size mismatch")
        if any([d == 0 for d in self.std]):
            raise AssertionError("normalization std must be non-null")

    def __call__(self, sample):
        """Normalizes a given sample.

        Args:
            sample: the sample to normalize. If given as a PIL image, it
            will be converted to a numpy array first.

        Returns:
            The warped sample, in a numpy array of type ``self.out_type``.
        """
        if isinstance(sample, PIL.Image.Image):
            sample = np.asarray(sample)
        if isinstance(sample, np.ndarray):
            return ((sample - self.mean) / self.std).astype(self.out_type)
        elif isinstance(sample, torch.Tensor):
            out = torchvision.transforms.functional.normalize(sample,
                                                              torch.from_numpy(self.mean),
                                                              torch.from_numpy(self.std))
            if self.out_type == np.float32:
                return out.float()
            else:
                raise AssertionError("missing impl for non-float torch normalize output")
        else:
            raise AssertionError("sample type should be np.ndarray or torch.Tensor")

    def invert(self, sample):
        """Inverts the normalization."""
        if isinstance(sample, PIL.Image.Image):
            sample = np.asarray(sample)
        return (sample * self.std + self.mean).astype(self.out_type)

    def __repr__(self):
        """Provides print-friendly output for class attributes."""
        return self.__class__.__name__ + ": {{mean: {0}, std: {1}, out_type: {2}}}".format(self.mean, self.std, self.out_type)


class NormalizeMinMax(object):
    """Normalizes a given image using a set of minimum and maximum values.

    The samples will be transformed such that ``s = (s - min) / (max - min)``.

    Note that this operation is also not restricted to images.

    Attributes:
        min: an array of minimum values to subtract with.
        max: an array of maximum values to divide with.
        out_type: the output data type to cast the normalization result to.
    """

    def __init__(self, min, max, out_type=np.float32):
        """Validates and initializes normalization parameters.

        Args:
            min: an array of minimum values to subtract with.
            max: an array of maximum values to divide with.
            out_type: the output data type to cast the normalization result to.
        """
        self.out_type = out_type
        self.min = np.asarray(min).astype(out_type)
        if self.min.ndim == 0:
            self.min = np.expand_dims(self.min, 0)
        self.max = np.asarray(max).astype(out_type)
        if self.max.ndim == 0:
            self.max = np.expand_dims(self.max, 0)
        if self.min.ndim != 1 or self.max.ndim != 1:
            raise AssertionError("normalization params should be a 1-d array (one value per channel)")
        if self.min.size != self.max.size:
            raise AssertionError("normalization params size mismatch")
        self.diff = self.max - self.min
        if any([d == 0 for d in self.diff]):
            raise AssertionError("normalization diff must be non-null")

    def __call__(self, sample):
        """Normalizes a given sample.

        Args:
            sample: the sample to normalize. If given as a PIL image, it
            will be converted to a numpy array first.

        Returns:
            The warped sample, in a numpy array of type ``self.out_type``.
        """
        if isinstance(sample, PIL.Image.Image):
            sample = np.asarray(sample)
        elif not isinstance(sample, np.ndarray):
            raise AssertionError("sample type should be np.ndarray")
        return ((sample - self.min) / self.diff).astype(self.out_type)

    def invert(self, sample):
        """Inverts the normalization."""
        if isinstance(sample, PIL.Image.Image):
            sample = np.asarray(sample)
        return (sample * self.diff + self.min).astype(self.out_type)

    def __repr__(self):
        """Provides print-friendly output for class attributes."""
        return self.__class__.__name__ + ": {{min: {0}, max: {1}, out_type: {2}}}".format(self.min, self.max, self.out_type)



"""Samplers module.

This module contains classes used for raw dataset rebalancing or augmentation.

All samplers here should aim to be compatible with PyTorch's sampling interface
(torch.utils.data.sampler.Sampler) so that they can be instantiated at runtime
through a configuration file and used as the input of a data loader.
"""
import copy
import logging

import torch
import torch.utils.data.sampler

import thelper.data.utils

logger = logging.getLogger(__name__)


class WeightedSubsetRandomSampler(torch.utils.data.sampler.Sampler):
    r"""Provides a rebalanced list of sample indices to use in a data loader.

    Given a list of sample indices and the corresponding list of class labels, this sampler
    will produce a new list of indices that rebalances the distribution of samples according
    to a specified strategy. It can also optionally scale the dataset's total sample count to
    avoid undersampling large classes as smaller ones get bigger.

    The currently implemented strategies are:

      * ``random``: will return a list of randomly picked samples based on the multinomial \
        distribution of the initial class weights. This sampling is done with replacement, \
        meaning that each index is picked independently of the already-picked ones.

      * ``uniform``: will rebalance the dataset by normalizing the sample count of all classes, \
        oversampling and undersampling as required to distribute all samples equally. All \
        removed or duplicated samples are selected randomly without replacement.

      * ``root``: will rebalance the dataset by normalizing class weight using an n-th degree \
        root. More specifically, for a list of initial class weights :math:`W^0=\{w_1^0, w_2^0, ... w_n^0\}`, \
        we compute the adjusted weight :math:`w_i` of each class via:

        .. math::
          w_i = \frac{\sqrt[\leftroot{-1}\uproot{3}n]{w_i^0}}{\sum_j\sqrt[\leftroot{-1}\uproot{3}n]{w_j^0}}

        Then, according to the new distribution of weights, all classes are oversampled and
        undersampled as required to reobtain the dataset's total sample count (which may be
        scaled). All removed or duplicated samples are selected randomly without replacement.

        Note that with the ``root`` strategy, if a very large root degree ``n`` is used, this
        strategy is equivalent to ``uniform``. The ``root`` strategy essentially provides a
        solution for extremely unbalanced datasets where uniform oversampling and undersampling
        would be too aggressive.

    By default, this interface will try to keep the dataset size constant and balance oversampling
    with undersampling. If undersampling is undesired, the user can increase the total dataset
    size via a scale factor. Finally, note that the rebalanced list of indices is generated by
    this interface every time the ``__iter__`` function is called, meaning two consecutive lists
    might not contain the exact same indices.

    Example configuration file::

        # ...
        # the sampler is defined inside the 'loaders' field
        "loaders": {
            # ...
            # this field is completely optional, and can be omitted entirely
            "sampler": {
                # the type of the sampler we want to instantiate
                "type": "thelper.data.samplers.WeightedSubsetRandomSampler",
                # the parameters passed to the sampler's constructor
                "params": {
                    "stype": "root3",
                    "scale": 1.2
                },
                # specifies whether the sampler should receive class labels
                "pass_labels": true
            },
            # ...
        },
        # ...

    Attributes:
        nb_samples: total number of samples to rebalance (i.e. scaled size of original dataset).
        label_groups: map that splits all samples indices into groups based on labels.
        stype: name of the rebalancing strategy to use.
        indices: copy of the original list of sample indices provided in the constructor.
        sample_weights: list of weights used for random sampling.
        label_counts: number of samples in each class for the ``uniform`` and ``root`` strategies.
        seeds: dictionary of seeds to use when initializing RNG state.
        epoch: epoch number used to reinitialize the RNG to an epoch-specific state.

    .. seealso::
        | :func:`thelper.data.utils.create_loaders`
        | :func:`thelper.data.utils.get_class_weights`
    """

    def __init__(self, indices, labels, stype="uniform", scale=1.0, seeds=None, epoch=0):
        """Receives sample indices, labels, rebalancing strategy, and dataset scaling factor.

        This function will validate all input arguments, parse and categorize samples according to
        labels, initialize rebalancing parameters, and determine sample counts for each valid class.
        Note that an empty list of indices is an acceptable input; the resulting object will also
        create and empty list of samples when ``__iter__`` is called.

        Args:
            indices: list of integers representing the indices of samples of interest in the dataset.
            labels: list of labels tied to the list of indices (must be the same length).
            stype: rebalancing strategy given as a string. Should be either "random", "uniform", or
                "rootX", where the 'X' is the degree to use in the root computation (float).
            scale: scaling factor used to increase/decrease the final number of sample indices to
                generate while rebalancing.
            seeds: dictionary of seeds to use when initializing RNG state.
            epoch: epoch number used to reinitialize the RNG to an epoch-specific state.
        """
        super().__init__(None)
        if not isinstance(indices, list) or not isinstance(labels, list):
            raise AssertionError("expected indices and labels to be provided as lists")
        if len(indices) != len(labels):
            raise AssertionError("mismatched indices/labels list sizes")
        if not isinstance(scale, float) or scale < 0:
            raise AssertionError("invalid scale parameter; should be greater than zero")
        self.seeds = {}
        if seeds is not None:
            if not isinstance(seeds, dict):
                raise AssertionError("unexpected seed pack type")
            self.seeds = seeds
        if not isinstance(epoch, int) or epoch < 0:
            raise AssertionError("invalid epoch value")
        self.epoch = epoch
        self.nb_samples = int(round(len(indices) * scale))
        if self.nb_samples > 0:
            self.stype = stype
            self.indices = copy.deepcopy(indices)
            self.label_groups = {}
            for idx, label in enumerate(labels):
                if label in self.label_groups:
                    self.label_groups[label].append(indices[idx])
                else:
                    self.label_groups[label] = [indices[idx]]
            if not isinstance(stype, str) or (stype not in ["uniform", "random"] and "root" not in stype):
                raise AssertionError("unexpected sampling type")
            if stype == "random":
                self.sample_weights = [1.0 / len(self.label_groups[label]) for label in labels]
            else:
                weights = thelper.data.utils.get_class_weights(self.label_groups, stype, invmax=False)
                self.label_counts = {}
                curr_nb_samples, max_sample_label = 0, None
                for label_idx, (label, indices) in enumerate(self.label_groups.items()):
                    self.label_counts[label] = int(self.nb_samples * weights[label])
                    curr_nb_samples += self.label_counts[label]
                    if max_sample_label is None or len(self.label_groups[label]) > len(self.label_groups[max_sample_label]):
                        max_sample_label = label
                if curr_nb_samples != self.nb_samples:
                    self.label_counts[max_sample_label] += self.nb_samples - curr_nb_samples

    def set_epoch(self, epoch=0):
        """Sets the current epoch number in order to offset the RNG state for sampling."""
        if not isinstance(epoch, int) or epoch < 0:
            raise AssertionError("invalid epoch value")
        self.epoch = epoch

    def __iter__(self):
        """Returns the list of rebalanced sample indices to load.

        Note that the indices are repicked every time this function is called, meaning that samples
        eliminated due to undersampling (or duplicated due to oversampling) might not receive the same
        treatment twice.

        This function will reseed the RNGs it uses every time it is called, and revert their state before
        returning its output.
        """
        if self.nb_samples == 0:
            self.epoch += 1
            return iter([])
        rng_state = None
        if "torch" in self.seeds:
            rng_state = torch.random.get_rng_state()
            torch.random.manual_seed(self.seeds["torch"] + self.epoch)
        result = None
        if self.stype == "random":
            result = (self.indices[idx] for idx in torch.multinomial(
                torch.FloatTensor(self.sample_weights),self.nb_samples, replacement=True))
        elif self.stype == "uniform" or "root" in self.stype:
            indices = []
            for label, count in self.label_counts.items():
                max_samples = len(self.label_groups[label])
                while count > 0:
                    subidxs = torch.randperm(max_samples)
                    for subidx in range(min(count, max_samples)):
                        indices.append(self.label_groups[label][subidxs[subidx]])
                    count -= max_samples
            if len(indices) != self.nb_samples:
                raise AssertionError("messed up something internally...")
            result = (indices[i] for i in torch.randperm(len(indices)))
        if rng_state is not None:
            torch.random.set_rng_state(rng_state)
        self.epoch += 1
        return result

    def __len__(self):
        """Returns the number of sample indices that will be generated by this interface.

        This number is the scaled size of the originally provided sample indices list.
        """
        return self.nb_samples


class SubsetRandomSampler(torch.utils.data.sampler.Sampler):
    r"""Samples elements randomly from a given list of indices, without replacement.

    This specialization handles seeding based on the epoch number, and scaling (via duplication/decimation)
    of samples.

    Arguments:
        indices (list): a list of indices
        seeds (dict): dictionary of seeds to use when initializing RNG state.
        epoch (int): epoch number used to reinitialize the RNG to an epoch-specific state.
        scale (float): scaling factor used to increase/decrease the final number of samples.
    """

    def __init__(self, indices, seeds=None, epoch=0, scale=1.0):
        super().__init__(indices)
        self.seeds = {}
        if seeds is not None:
            if not isinstance(seeds, dict):
                raise AssertionError("unexpected seed pack type")
            self.seeds = seeds
        if not isinstance(epoch, int) or epoch < 0:
            raise AssertionError("invalid epoch value")
        self.epoch = epoch
        self.indices = indices
        if not isinstance(scale, (float, int)) or scale < 0:
            raise AssertionError("invalid scale parameter; should be greater than zero")
        self.num_samples = int(round(len(self.indices) * scale))

    def set_epoch(self, epoch=0):
        """Sets the current epoch number in order to offset the RNG state for sampling."""
        if not isinstance(epoch, int) or epoch < -1:
            raise AssertionError("invalid epoch value")
        self.epoch = epoch

    def __iter__(self):
        rng_state = None
        if "torch" in self.seeds:
            rng_state = torch.random.get_rng_state()
            torch.random.manual_seed(self.seeds["torch"] + self.epoch)
        indices = []
        max_samples = len(self.indices)
        req_count = self.num_samples
        while req_count > 0:
            subidxs = torch.randperm(max_samples)
            for subidx in range(min(req_count, max_samples)):
                indices.append(self.indices[subidxs[subidx]])
            req_count -= max_samples
        result = (indices[i] for i in torch.randperm(len(indices)))
        if rng_state is not None:
            torch.random.set_rng_state(rng_state)
        self.epoch += 1
        return result

    def __len__(self):
        return self.num_samples


class SubsetSequentialSampler(torch.utils.data.sampler.Sampler):
    r"""Samples element indices sequentially, always in the same order.

    Arguments:
        indices (list): a list of indices
    """

    def __init__(self, indices):
        super().__init__(indices)
        self.indices = indices

    def __iter__(self):
        return iter(self.indices)

    def __len__(self):
        return len(self.indices)
